# -*- coding: utf-8 -*-
#
#    BitcoinLib - Python Cryptocurrency Library
#    BitGo Client
#    © 2017-2019 July - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
from datetime import datetime
from bitcoinlib.main import MAX_TRANSACTIONS
from bitcoinlib.services.baseclient import BaseClient, ClientError
from bitcoinlib.transactions import Transaction

_logger = logging.getLogger(__name__)

PROVIDERNAME = 'bitgo'
LIMIT_TX = 49


class BitGoClient(BaseClient):

    def __init__(self, network, base_url, denominator, *args):
        super(self.__class__, self).__init__(network, PROVIDERNAME, base_url, denominator, *args)

    def compose_request(self, category, data, cmd='', variables=None, method='get'):
        if data:
            data = '/' + data
        url_path = category + data
        if cmd != '':
            url_path += '/' + cmd
        return self.request(url_path, variables, method=method)

    def getbalance(self, addresslist):
        balance = 0
        for address in addresslist:
            res = self.compose_request('address', address)
            balance += res['balance']
        return balance

    def getutxos(self, address, after_txid='', limit=MAX_TRANSACTIONS):
        utxos = []
        skip = 0
        total = 1
        while total > skip:
            variables = {'limit': 100, 'skip': skip}
            res = self.compose_request('address', address, 'unspents', variables)
            for utxo in res['unspents'][::-1]:
                if utxo['tx_hash'] == after_txid:
                    break
                utxos.append(
                    {
                        'address': utxo['address'],
                        'txid': utxo['tx_hash'],
                        'confirmations': utxo['confirmations'],
                        'output_n': utxo['tx_output_n'],
                        'input_n': 0,
                        'block_height': utxo['blockHeight'],
                        'fee': None,
                        'size': 0,
                        'value': int(round(utxo['value'] * self.units, 0)),
                        'script': utxo['script'],
                        'date': datetime.strptime(utxo['date'], "%Y-%m-%dT%H:%M:%S.%fZ")
                     }
                )
            total = res['total']
            skip = res['start'] + res['count']
            if skip > 2000:
                _logger.info("BitGoClient: UTXO's list has been truncated, list is incomplete")
                break
        return utxos[::-1][:limit]

    # RAW TRANSACTION DOES NOT CONTAIN CORRECT RAW TRANSACTION (MISSING SIGS)
    # def gettransaction(self, txid):
    #     tx = self.compose_request('tx', txid)
    #     t = Transaction.parse_hex(tx['hex'], strict=False, network=self.network)
    #     t.status = 'unconfirmed'
    #     t.date = None
    #     if tx['confirmations']:
    #         t.status = 'confirmed'
    #         t.date = datetime.strptime(tx['date'], "%Y-%m-%dT%H:%M:%S.%fZ")
    #     t.confirmations = tx['confirmations']
    #     if 'height' in tx:
    #         t.block_height = tx['height']
    #         t.block_hash = tx['blockhash']
    #     t.fee = tx['fee']
    #     t.rawtx = to_bytes(tx['hex'])
    #     t.size = len(tx['hex']) // 2
    #     t.network = self.network
    #     if t.coinbase:
    #         input_values = []
    #         t.input_total = t.output_total
    #     else:
    #         input_values = [(inp['account'], -inp['value']) for inp in tx['entries'] if inp['value'] < 0]
    #         if len(input_values) >= 49:
    #             raise ClientError("More then 49 transaction inputs not supported by bitgo")
    #         t.input_total = sum([x[1] for x in input_values])
    #     for i in t.inputs:
    #         if not i.address and not t.coinbase:
    #             raise ClientError("Address missing in input. Provider might not support segwit transactions")
    #         if len(t.inputs) != len(input_values):
    #             i.value = None
    #             continue
    #         value = [x[1] for x in input_values if x[0] == i.address]
    #         if len(value) != 1:
    #             _logger.info("BitGoClient: Address %s input value should be found exactly 1 times in value list" %
    #                             i.address)
    #             i.value = None
    #         else:
    #             i.value = value[0]
    #     for o in t.outputs:
    #         o.spent = None
    #     if t.input_total != t.output_total + t.fee:
    #         t.input_total = t.output_total + t.fee
    #     return t

    # RAW TRANSACTION DOES NOT CONTAIN CORRECT RAW TRANSACTION (MISSING SIGS)
    # def gettransactions(self, address, after_txid='', limit=MAX_TRANSACTIONS):
    #     txs = []
    #     txids = []
    #     skip = 0
    #     total = 1
    #     while total > skip:
    #         variables = {'limit': LIMIT_TX, 'skip': skip}
    #         res = self.compose_request('address', address, 'tx', variables)
    #         for tx in res['transactions']:
    #             if tx['id'] not in txids:
    #                 txids.insert(0, tx['id'])
    #         total = res['total']
    #         # if total > 2000:
    #         #     raise ClientError("BitGoClient: Transactions list limit exceeded > 2000")
    #         skip = res['start'] + res['count']
    #         if len(txids) > limit:
    #             break
    #     if after_txid:
    #         txids = txids[txids.index(after_txid) + 1:]
    #     for txid in txids[:limit]:
    #         txs.append(self.gettransaction(txid))
    #     return txs

    # RAW TRANSACTION DOES NOT CONTAIN CORRECT RAW TRANSACTION (MISSING SIGS)
    # def getrawtransaction(self, txid):
    #     tx = self.compose_request('tx', txid)
    #     t = Transaction.parse_hex(tx['hex'], strict=False, network=self.network)
    #     for i in t.inputs:
    #         if not i.address:
    #             raise ClientError("Address missing in input. Provider might not support segwit transactions")
    #     return tx['hex']

    # def sendrawtransaction

    def estimatefee(self, blocks):
        res = self.compose_request('tx', 'fee', variables={'numBlocks': blocks})
        return res['feePerKb']

    def blockcount(self):
        return self.compose_request('block', 'latest')['height']

    # def mempool

    # def getblock(self, blockid, parse_transactions, page, limit):
    #     bd = self.compose_request('block', str(blockid))
    #     if parse_transactions:
    #         txs = []
    #         for txid in bd['transactions'][(page-1)*limit:page*limit]:
    #             try:
    #                 txs.append(self.gettransaction(txid))
    #             except Exception as e:
    #                 _logger.error("Could not parse tx %s with error %s" % (txid, e))
    #     else:
    #         txs = bd['transactions']
    #
    #     block = {
    #         'bits': None,
    #         'depth': None,
    #         'hash': bd['id'],
    #         'height': bd['height'],
    #         'merkle_root': bd['merkleRoot'],
    #         'nonce': bd['nonce'],
    #         'prev_block': bd['previous'],
    #         'time': datetime.strptime(bd['date'], "%Y-%m-%dT%H:%M:%S.%fZ").replace(microsecond=0),
    #         'total_txs': len(bd['transactions']),
    #         'txs': txs,
    #         'version': bd['version'],
    #         'page': page,
    #         'pages': None if not limit else int(len(bd['transactions']) // limit) + (len(bd['transactions']) % limit > 0),
    #         'limit': limit
    #     }
    #     return block

    # def isspent(self, txid, index):

    # def getinfo(self):



# -*- coding: utf-8 -*-
#
#    BitcoinLib - Python Cryptocurrency Library
#    Unit Tests for Database
#    © 2019 December - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest
from sqlalchemy.exc import OperationalError
from tests.db_0_5 import Db as DbInitOld
from bitcoinlib.db import *
from bitcoinlib.db_cache import *
from bitcoinlib.wallets import Wallet, WalletError
from bitcoinlib.services.services import Service


DATABASEFILE_UNITTESTS = os.path.join(str(BCL_DATABASE_DIR), 'bitcoinlib.unittest.sqlite')
DATABASEFILE_TMP = os.path.join(str(BCL_DATABASE_DIR), 'bitcoinlib.tmp.sqlite')
DATABASEFILE_CACHE_TMP = os.path.join(str(BCL_DATABASE_DIR), 'bitcoinlib_cache.tmp.sqlite')


class TestDb(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        if os.path.isfile(DATABASEFILE_TMP):
            os.remove(DATABASEFILE_TMP)
        if os.path.isfile(DATABASEFILE_CACHE_TMP):
            os.remove(DATABASEFILE_CACHE_TMP)

    def test_database_upgrade(self):
        if os.path.isfile(DATABASEFILE_UNITTESTS):
            os.remove(DATABASEFILE_UNITTESTS)
        dbold = DbInitOld(DATABASEFILE_UNITTESTS)

        # self.assertFalse('latest_txid' in dbold.engine.execute("SELECT * FROM keys").keys())
        # self.assertFalse('address' in dbold.engine.execute("SELECT * FROM transaction_inputs").keys())
        # version_db = dbold.session.query(DbConfig.value).filter_by(variable='version').scalar()
        # self.assertEqual(version_db, '0.4.10')

    def test_database_create_drop(self):
        dbtmp = Db(DATABASEFILE_TMP)
        Wallet.create("tmpwallet", db_uri=DATABASEFILE_TMP)
        self.assertRaisesRegexp(WalletError, "Wallet with name 'tmpwallet' already exists",
                                Wallet.create, 'tmpwallet', db_uri=DATABASEFILE_TMP)
        dbtmp.drop_db(yes_i_am_sure=True)
        Wallet.create("tmpwallet", db_uri=DATABASEFILE_TMP)

    def test_database_cache_create_drop(self):
        dbtmp = DbCache(DATABASEFILE_CACHE_TMP)
        srv = Service(cache_uri=DATABASEFILE_CACHE_TMP, exclude_providers=['bitaps', 'bitgo'])
        t = srv.gettransaction('68104dbd6819375e7bdf96562f89290b41598df7b002089ecdd3c8d999025b13')
        if t:
            self.assertGreaterEqual(srv.results_cache_n, 0)
            srv.gettransaction('68104dbd6819375e7bdf96562f89290b41598df7b002089ecdd3c8d999025b13')
            self.assertGreaterEqual(srv.results_cache_n, 1)
            dbtmp.drop_db()
            self.assertRaisesRegex(OperationalError, "", srv.gettransaction,
                                   '68104dbd6819375e7bdf96562f89290b41598df7b002089ecdd3c8d999025b13')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
#
#    BitcoinLib - Python Cryptocurrency Library
#    Cache DataBase - SqlAlchemy database definitions for caching
#    © 2020 February - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from sqlalchemy import create_engine
from sqlalchemy import Column, Integer, BigInteger, String, Boolean, ForeignKey, DateTime, Enum, LargeBinary
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship, close_all_sessions
# try:
#     import mysql.connector
#     from parameterized import parameterized_class
#     import psycopg2
#     from psycopg2 import sql
#     from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
# except ImportError as e:
#     print("Could not import all modules. Error: %s" % e)
#     # from psycopg2cffi import compat  # Use for PyPy support
#     # compat.register()
#     pass  # Only necessary when mysql or postgres is used
from urllib.parse import urlparse
from bitcoinlib.main import *


_logger = logging.getLogger(__name__)
try:
    dbcacheurl_obj = urlparse(DEFAULT_DATABASE_CACHE)
    if dbcacheurl_obj.netloc:
        dbcacheurl = dbcacheurl_obj.netloc.replace(dbcacheurl_obj.password, 'xxx')
    else:
        dbcacheurl = dbcacheurl_obj.path
    _logger.info("Default Cache Database %s" % dbcacheurl)
except Exception:
    _logger.warning("Default Cache Database: unable to parse URL")
Base = declarative_base()


class WitnessTypeTransactions(enum.Enum):
    legacy = "legacy"
    segwit = "segwit"


class DbCache:
    """
    Cache Database object. Initialize database and open session when creating database object.

    Create new database if is doesn't exist yet

    """
    def __init__(self, db_uri=None):
        self.engine = None
        self.session = None
        if db_uri is None:
            db_uri = DEFAULT_DATABASE_CACHE
        elif not db_uri:
            return
        self.o = urlparse(db_uri)

        # if self.o.scheme == 'mysql':
        #     raise Warning("Could not connect to cache database. MySQL databases not supported at the moment, "
        #                   "because bytes strings are not supported as primary keys")

        if not self.o.scheme or len(self.o.scheme) < 2:
            db_uri = 'sqlite:///%s' % db_uri
        if db_uri.startswith("sqlite://") and ALLOW_DATABASE_THREADS:
            db_uri += "&" if "?" in db_uri else "?"
            db_uri += "check_same_thread=False"
        if self.o.scheme == 'mysql':
            db_uri += "&" if "?" in db_uri else "?"
            db_uri += 'binary_prefix=true'
        self.engine = create_engine(db_uri, isolation_level='READ UNCOMMITTED')

        Session = sessionmaker(bind=self.engine)
        Base.metadata.create_all(self.engine)
        self.db_uri = db_uri
        _logger.info("Using cache database: %s://%s:%s/%s" % (self.o.scheme or '', self.o.hostname or '',
                                                              self.o.port or '', self.o.path or ''))
        self.session = Session()

    def drop_db(self):
        self.session.commit()
        self.session.close_all()
        close_all_sessions()
        Base.metadata.drop_all(self.engine)


class DbCacheTransactionNode(Base):
    """
    Link table for cache transactions and addresses
    """
    __tablename__ = 'cache_transactions_node'
    txid = Column(LargeBinary(32), ForeignKey('cache_transactions.txid'), primary_key=True)
    transaction = relationship("DbCacheTransaction", back_populates='nodes', doc="Related transaction object")
    index_n = Column(Integer, primary_key=True, doc="Order of input/output in this transaction")
    value = Column(BigInteger, default=0, doc="Value of transaction input")
    address = Column(String(255), index=True, doc="Address string base32 or base58 encoded")
    script = Column(LargeBinary, doc="Locking or unlocking script")
    witnesses = Column(LargeBinary, doc="Witnesses (signatures) used in Segwit transaction inputs")
    sequence = Column(BigInteger, default=0xffffffff,
                      doc="Transaction sequence number. Used for timelock transaction inputs")
    is_input = Column(Boolean, primary_key=True, doc="True if input, False if output")
    spent = Column(Boolean, default=None, doc="Is output spent?")
    ref_txid = Column(LargeBinary(32), index=True, doc="Transaction hash of input which spends this output")
    ref_index_n = Column(BigInteger, doc="Index number of transaction input which spends this output")

    def prev_txid(self):
        if self.is_input:
            return self.ref_txid

    def output_n(self):
        if self.is_input:
            return self.ref_index_n

    def spending_txid(self):
        if not self.is_input:
            return self.ref_txid

    def spending_index_n(self):
        if not self.is_input:
            return self.ref_index_n


class DbCacheTransaction(Base):
    """
    Transaction Cache Table

    Database which stores transactions received from service providers as cache

    """
    __tablename__ = 'cache_transactions'
    txid = Column(LargeBinary(32), primary_key=True, doc="Hexadecimal representation of transaction hash or transaction ID")
    date = Column(DateTime, doc="Date when transaction was confirmed and included in a block")
    version = Column(BigInteger, default=1,
                     doc="Tranaction version. Default is 1 but some wallets use another version number")
    locktime = Column(BigInteger, default=0,
                      doc="Transaction level locktime. Locks the transaction until a specified block "
                          "(value from 1 to 5 million) or until a certain time (Timestamp in seconds after 1-jan-1970)."
                          " Default value is 0 for transactions without locktime")
    confirmations = Column(Integer, default=0,
                           doc="Number of confirmation when this transaction is included in a block. "
                               "Default is 0: unconfirmed")
    block_height = Column(Integer, index=True, doc="Height of block this transaction is included in")
    network_name = Column(String(20), doc="Blockchain network name of this transaction")
    fee = Column(BigInteger, doc="Transaction fee")
    nodes = relationship("DbCacheTransactionNode", cascade="all,delete",
                         doc="List of all inputs and outputs as DbCacheTransactionNode objects")
    order_n = Column(Integer, doc="Order of transaction in block")
    witness_type = Column(Enum(WitnessTypeTransactions), default=WitnessTypeTransactions.legacy,
                          doc="Transaction type enum: legacy or segwit")


class DbCacheAddress(Base):
    """
    Address Cache Table

    Stores transactions and unspent outputs (UTXO's) per address

    """
    __tablename__ = 'cache_address'
    address = Column(String(255), primary_key=True, doc="Address string base32 or base58 encoded")
    network_name = Column(String(20), doc="Blockchain network name of this transaction")
    balance = Column(BigInteger, default=0, doc="Total balance of UTXO's linked to this key")
    last_block = Column(Integer, doc="Number of last updated block")
    last_txid = Column(LargeBinary(32), doc="Transaction ID of latest transaction in cache")
    n_utxos = Column(Integer, doc="Total number of UTXO's for this address")
    n_txs = Column(Integer, doc="Total number of transactions for this address")


class DbCacheBlock(Base):
    """
    Block Cache Table

    Stores block headers
    """
    __tablename__ = 'cache_blocks'
    height = Column(Integer, primary_key=True, doc="Height or sequence number for this block")
    block_hash = Column(LargeBinary(32), index=True, doc="Hash of this block")
    network_name = Column(String(20), doc="Blockchain network name")
    version = Column(BigInteger, doc="Block version to specify which features are used (hex)")
    prev_block = Column(LargeBinary(32), doc="Block hash of previous block")
    merkle_root = Column(LargeBinary(32), doc="Merkle root used to validate transaction in block")
    time = Column(BigInteger, doc="Timestamp to indicated when block was created")
    bits = Column(BigInteger, doc="Encoding for proof-of-work, used to determine target and difficulty")
    nonce = Column(BigInteger, doc="Nonce (number used only once or n-once) is used to create different block hashes")
    tx_count = Column(Integer, doc="Number of transactions included in this block")


class DbCacheVars(Base):
    """
    Table to store various blockchain related variables
    """
    __tablename__ = 'cache_variables'
    varname = Column(String(50), primary_key=True, doc="Variable unique name")
    network_name = Column(String(20), primary_key=True, doc="Blockchain network name of this transaction")
    value = Column(String(255), doc="Value of variable")
    type = Column(String(20), doc="Type of variable: int, string or float")
    expires = Column(DateTime, doc="Datetime value when variable expires")



# -*- coding: utf-8 -*-
#
#    BitcoinLib - Python Cryptocurrency Library
#    Smartbit.com.au client
#    © 2019 August - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
from datetime import datetime
from bitcoinlib.main import MAX_TRANSACTIONS
from bitcoinlib.services.baseclient import BaseClient
from bitcoinlib.transactions import Transaction
from bitcoinlib.encoding import varstr
from bitcoinlib.keys import Address

_logger = logging.getLogger(__name__)

PROVIDERNAME = 'smartbit'
REQ_LIMIT = 10
REQ_LIMIT_TOTAL = 50
# Please note: In the Bitaps API, the first couple of Bitcoin blocks are not correctly indexed,
# so transactions from these blocks are missing.


class SmartbitClient(BaseClient):

    def __init__(self, network, base_url, denominator, *args):
        super(self.__class__, self).__init__(network, PROVIDERNAME, base_url, denominator, *args)

    def compose_request(self, category, command='', data='', variables=None, req_type='blockchain', method='get'):
        url_path = req_type + '/' + category
        if data:
            if url_path[-1:] != '/':
                url_path += '/'
            url_path += data
        if command:
            url_path += '/' + command
        return self.request(url_path, variables=variables, method=method)

    def _parse_transaction(self, tx):
        status = 'unconfirmed'
        if tx['confirmations']:
            status = 'confirmed'
        witness_type = 'legacy'
        if 'inputs' in tx and [ti['witness'] for ti in tx['inputs'] if ti['witness'] and ti['witness'] != ['NULL']]:
            witness_type = 'segwit'
        input_total = tx['input_amount_int']
        t_time = None
        if tx['time']:
            t_time = datetime.utcfromtimestamp(tx['time'])
        t = Transaction(locktime=tx['locktime'], version=int(tx['version']), network=self.network, fee=tx['fee_int'],
                        size=tx['size'], txid=tx['txid'], date=t_time,
                        confirmations=tx['confirmations'], block_height=tx['block'], status=status,
                        input_total=input_total, coinbase=tx['coinbase'],
                        output_total=tx['output_amount_int'], witness_type=witness_type)
        index_n = 0
        if tx['coinbase']:
            t.add_input(prev_txid=b'\00' * 32, output_n=0, value=0)
        else:
            for ti in tx['inputs']:
                unlocking_script = ti['script_sig']['hex']
                witness_type = 'legacy'
                if ti['witness'] and ti['witness'] != ['NULL']:
                    address = Address.parse(ti['addresses'][0])
                    if address.script_type == 'p2sh':
                        witness_type = 'p2sh-segwit'
                    else:
                        witness_type = 'segwit'
                t.add_input(prev_txid=ti['txid'], output_n=ti['vout'], unlocking_script=unlocking_script,
                            index_n=index_n, value=ti['value_int'], address=ti['addresses'][0], sequence=ti['sequence'],
                            witness_type=witness_type, witnesses=ti['witness'] if ti['witness'] != ['NULL'] else [],
                            strict=False)
                index_n += 1

        for to in tx['outputs']:
            spent = False
            spending_txid = None
            if 'spend_txid' in to and to['spend_txid']:
                spent = True
                spending_txid = to['spend_txid']
            address = ''
            if to['addresses']:
                address = to['addresses'][0]
            t.add_output(value=to['value_int'], address=address, lock_script=to['script_pub_key']['hex'],
                         spent=spent, output_n=to['n'], spending_txid=spending_txid, strict=False)
        return t

    def getbalance(self, addresslist):
        res = self.compose_request('address', 'wallet', ','.join(addresslist))
        return res['wallet']['total']['balance_int']

    def getutxos(self, address, after_txid='', limit=MAX_TRANSACTIONS):
        utxos = []
        utxo_list = []
        next_link = ''
        while True:
            variables = {'limit': REQ_LIMIT, 'next': next_link, 'dir': 'asc'}
            res = self.compose_request('address', 'unspent', address, variables=variables)
            next_link = res['paging']['next']
            for utxo in res['unspent']:
                utxo_list.append(utxo['txid'])
                if utxo['txid'] == after_txid:
                    utxo_list = []
            if not next_link or len(utxos) > REQ_LIMIT_TOTAL:
                break
        for txid in utxo_list[:limit]:
            t = self.gettransaction(txid)
            for utxo in t.outputs:
                if utxo.address != address:
                    continue
                utxos.append(
                    {
                        'address': utxo.address,
                        'txid': t.txid,
                        'confirmations': t.confirmations,
                        'output_n': utxo.output_n,
                        'input_n': 0,
                        'block_height': t.block_height,
                        'fee': t.fee,
                        'size': t.size,
                        'value': utxo.value,
                        'script': utxo.lock_script.hex(),
                        'date': t.date
                    })
        return utxos

    def gettransaction(self, txid):
        res = self.compose_request('tx', data=txid)
        return self._parse_transaction(res['transaction'])

    def gettransactions(self, address, after_txid='', limit=MAX_TRANSACTIONS):
        txs = []
        next_link = ''
        while True:
            variables = {'limit': REQ_LIMIT, 'next': next_link, 'dir': 'asc'}
            res = self.compose_request('address', data=address, variables=variables)
            next_link = '' if 'transaction_paging' not in res['address'] else \
                res['address']['transaction_paging']['next']
            if 'transactions' not in res['address']:
                break
            res_tx = sorted(res['address']['transactions'], key=lambda k: (k['block'] is None, k['block']))
            for tx in res_tx:
                t = self._parse_transaction(tx)
                txs.append(t)
                if t.txid == after_txid:
                    txs = []
            if not next_link or len(txs) > REQ_LIMIT_TOTAL:
                break
        return txs[:limit]

    def getrawtransaction(self, txid):
        res = self.compose_request('tx', data=txid, command='hex')
        return res['hex'][0]['hex']

    def sendrawtransaction(self, rawtx):
        res = self.compose_request('pushtx', variables={'hex': rawtx}, method='post')
        return {
            'txid': res['txid'],
            'response_dict': res
        }

    # def estimatefee

    def blockcount(self):
        return self.compose_request('totals')['totals']['block_count'] - 1

    def mempool(self, txid):
        if txid:
            tx = self.compose_request('tx', data=txid)
            if tx['transaction']['confirmations'] == 0:
                return [tx['transaction']['hash']]
        return []

    def getblock(self, blockid, parse_transactions, page, limit):
        if limit > 100:
            limit = 100
        if page > 1:  # Paging does not work with Smartbit
            return False
        variables = {'limit': limit}
        bd = self.compose_request('block', str(blockid), variables=variables)['block']
        if parse_transactions:
            txs = []
            for tx in bd['transactions']:
                # try:
                txs.append(self._parse_transaction(tx))
                # except Exception as e:
                #     _logger.error("Could not parse tx %s with error %s" % (tx['txid'], e))
        else:
            txs = [tx['txid'] for tx in bd['transactions']]

        block = {
            'bits': int(bd['bits'], 16),
            'depth': bd['confirmations'],
            'block_hash': bd['hash'],
            'height': bd['height'],
            'merkle_root': bd['merkleroot'],
            'nonce': bd['nonce'],
            'prev_block': bd['previous_block_hash'],
            'time': bd['time'],
            'tx_count': bd['transaction_count'],
            'txs': txs,
            'version': bd['version'],
            'page': page,
            'pages': None if not limit else int(bd['transaction_count'] // limit) + (bd['transaction_count'] % limit > 0),
            'limit': limit
        }
        return block

    def isspent(self, txid, output_n):
        t = self.gettransaction(txid)
        return 1 if t.outputs[output_n].spent else 0

    # def getinfo(self):



# -*- coding: utf-8 -*-
#
#    BitcoinLib - Python Cryptocurrency Library
#    CryptoID Chainz client
#    © 2018-2019 July - 1200 Web Development <http://1200wd.com/>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging
from datetime import datetime
from bitcoinlib.main import MAX_TRANSACTIONS
from bitcoinlib.services.baseclient import BaseClient, ClientError
from bitcoinlib.transactions import Transaction
from bitcoinlib.encoding import to_bytes


_logger = logging.getLogger(__name__)

PROVIDERNAME = 'cryptoid'


class CryptoID(BaseClient):

    def __init__(self, network, base_url, denominator, *args):
        super(self.__class__, self).__init__(network, PROVIDERNAME, base_url, denominator, *args)

    def compose_request(self, func=None, path_type='api', variables=None, method='get'):
        # API path: http://chainz.cryptoid.info/ltc/api.dws
        # Explorer path for raw tx: https://chainz.cryptoid.info/explorer/tx.raw.dws
        if variables is None:
            variables = {}
        if path_type == 'api':
            url_path = '%s/api.dws' % self.provider_coin_id
            variables.update({'q': func})
        else:
            url_path = 'explorer/tx.raw.dws'
            variables.update({'coin': self.provider_coin_id})
        if not self.api_key:
            raise ClientError("Request a CryptoID API key before using this provider")
        variables.update({'key': self.api_key})
        return self.request(url_path, variables, method)

    def getbalance(self, addresslist):
        balance = 0.0
        addresslist = self._addresslist_convert(addresslist)
        for a in addresslist:
            res = self.compose_request('getbalance', variables={'a': a.address})
            balance += float(res)
        return int(balance * self.units)

    def getutxos(self, address, after_txid='', limit=MAX_TRANSACTIONS):
        if not self.api_key:
            raise ClientError("Method getutxos() is not available for CryptoID without API key")
        utxos = []
        address = self._address_convert(address)
        variables = {'active': address.address}
        res = self.compose_request('unspent', variables=variables)
        if len(res['unspent_outputs']) > 50:
            _logger.info("CryptoID: Large number of outputs for address %s, "
                         "UTXO list may be incomplete" % address.address)
        for utxo in res['unspent_outputs'][::-1]:
            if utxo['txid'] == after_txid:
                break
            utxos.append({
                'address': address.address_orig,
                'txid': utxo['txid'],
                'confirmations': utxo['confirmations'],
                'output_n': utxo['tx_output_n'] if 'tx_output_n' in utxo else utxo['tx_ouput_n'],
                'input_n': 0,
                'block_height': None,
                'fee': None,
                'size': 0,
                'value': int(utxo['value']),
                'script': utxo['script'],
                'date': None
            })
        return utxos[::-1][:limit]

    def gettransaction(self, txid):
        variables = {'id': txid, 'hex': None}
        tx = self.compose_request(path_type='explorer', variables=variables)
        t = Transaction.parse_hex(tx['hex'], strict=False, network=self.network)
        variables = {'t': txid}
        tx_api = self.compose_request('txinfo', path_type='api', variables=variables)
        for n, i in enumerate(t.inputs):
            if i.script_type != 'coinbase':
                i.value = int(round(tx_api['inputs'][n]['amount'] * self.units, 0))
            else:
                i.value = 0
                t.coinbase = True
        for n, o in enumerate(t.outputs):
            o.spent = None
        if tx['confirmations']:
            t.status = 'confirmed'
        else:
            t.status = 'unconfirmed'
        t.date = datetime.utcfromtimestamp(tx['time'])
        t.block_height = tx_api['block']
        t.block_hash = tx['blockhash']
        t.confirmations = tx['confirmations']
        t.rawtx = bytes.fromhex(tx['hex'])
        t.size = tx['size']
        t.network = self.network
        t.locktime = tx['locktime']
        t.version = tx['version'].to_bytes(4, 'big')
        t.output_total = int(round(tx_api['total_output'] * self.units, 0))
        t.input_total = t.output_total
        t.fee = 0
        if t.input_total:
            t.fee = t.input_total - t.output_total
        return t

    def gettransactions(self, address, after_txid='', limit=MAX_TRANSACTIONS):
        address = self._address_convert(address)
        txs = []
        txids = []
        variables = {'active': address.address, 'n': 100}
        res = self.compose_request('multiaddr', variables=variables)
        for tx in res['txs']:
            if tx['hash'] not in txids:
                txids.insert(0, tx['hash'])
        if after_txid:
            txids = txids[txids.index(after_txid) + 1:]
        for txid in txids[:limit]:
            t = self.gettransaction(txid)
            txs.append(t)
        return txs

    def getrawtransaction(self, txid):
        variables = {'id': txid, 'hex': None}
        tx = self.compose_request(path_type='explorer', variables=variables)
        return tx['hex']

    # def sendrawtransaction

    # def estimatefee

    def blockcount(self):
        r = self.compose_request('getblockcount', path_type='api')
        return r

    def mempool(self, txid):
        variables = {'id': txid, 'hex': None}
        tx = self.compose_request(path_type='explorer', variables=variables)
        if 'confirmations' not in tx:
            return [tx['txid']]
        return []

    # def getblock

    # def isspent

    # def getinfo(self):
